# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class DecoratedRuntimeEnv(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'ray_env_hash': 'str',
        'name': 'str',
        'working_dir': 'str',
        'py_modules': 'list[str]',
        'pip_packages': 'str',
        'conda_env_name': 'str',
        'conda_values': 'object',
        'env_variables': 'dict(str, str)',
        'created_at': 'datetime',
        'creator_id': 'str',
        'creator': 'MiniUser',
        'num_jobs': 'int',
        'num_named_actors': 'int'
    }

    attribute_map = {
        'id': 'id',
        'ray_env_hash': 'ray_env_hash',
        'name': 'name',
        'working_dir': 'working_dir',
        'py_modules': 'py_modules',
        'pip_packages': 'pip_packages',
        'conda_env_name': 'conda_env_name',
        'conda_values': 'conda_values',
        'env_variables': 'env_variables',
        'created_at': 'created_at',
        'creator_id': 'creator_id',
        'creator': 'creator',
        'num_jobs': 'num_jobs',
        'num_named_actors': 'num_named_actors'
    }

    def __init__(self, id=None, ray_env_hash=None, name=None, working_dir=None, py_modules=None, pip_packages=None, conda_env_name=None, conda_values=None, env_variables=None, created_at=None, creator_id=None, creator=None, num_jobs=None, num_named_actors=None, local_vars_configuration=None):  # noqa: E501
        """DecoratedRuntimeEnv - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._ray_env_hash = None
        self._name = None
        self._working_dir = None
        self._py_modules = None
        self._pip_packages = None
        self._conda_env_name = None
        self._conda_values = None
        self._env_variables = None
        self._created_at = None
        self._creator_id = None
        self._creator = None
        self._num_jobs = None
        self._num_named_actors = None
        self.discriminator = None

        self.id = id
        self.ray_env_hash = ray_env_hash
        if name is not None:
            self.name = name
        if working_dir is not None:
            self.working_dir = working_dir
        if py_modules is not None:
            self.py_modules = py_modules
        if pip_packages is not None:
            self.pip_packages = pip_packages
        if conda_env_name is not None:
            self.conda_env_name = conda_env_name
        if conda_values is not None:
            self.conda_values = conda_values
        if env_variables is not None:
            self.env_variables = env_variables
        self.created_at = created_at
        self.creator_id = creator_id
        self.creator = creator
        self.num_jobs = num_jobs
        self.num_named_actors = num_named_actors

    @property
    def id(self):
        """Gets the id of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The id of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this DecoratedRuntimeEnv.


        :param id: The id of this DecoratedRuntimeEnv.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def ray_env_hash(self):
        """Gets the ray_env_hash of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The ray_env_hash of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: str
        """
        return self._ray_env_hash

    @ray_env_hash.setter
    def ray_env_hash(self, ray_env_hash):
        """Sets the ray_env_hash of this DecoratedRuntimeEnv.


        :param ray_env_hash: The ray_env_hash of this DecoratedRuntimeEnv.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and ray_env_hash is None:  # noqa: E501
            raise ValueError("Invalid value for `ray_env_hash`, must not be `None`")  # noqa: E501

        self._ray_env_hash = ray_env_hash

    @property
    def name(self):
        """Gets the name of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The name of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name):
        """Sets the name of this DecoratedRuntimeEnv.


        :param name: The name of this DecoratedRuntimeEnv.  # noqa: E501
        :type: str
        """

        self._name = name

    @property
    def working_dir(self):
        """Gets the working_dir of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The working_dir of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: str
        """
        return self._working_dir

    @working_dir.setter
    def working_dir(self, working_dir):
        """Sets the working_dir of this DecoratedRuntimeEnv.


        :param working_dir: The working_dir of this DecoratedRuntimeEnv.  # noqa: E501
        :type: str
        """

        self._working_dir = working_dir

    @property
    def py_modules(self):
        """Gets the py_modules of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The py_modules of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: list[str]
        """
        return self._py_modules

    @py_modules.setter
    def py_modules(self, py_modules):
        """Sets the py_modules of this DecoratedRuntimeEnv.


        :param py_modules: The py_modules of this DecoratedRuntimeEnv.  # noqa: E501
        :type: list[str]
        """

        self._py_modules = py_modules

    @property
    def pip_packages(self):
        """Gets the pip_packages of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The pip_packages of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: str
        """
        return self._pip_packages

    @pip_packages.setter
    def pip_packages(self, pip_packages):
        """Sets the pip_packages of this DecoratedRuntimeEnv.


        :param pip_packages: The pip_packages of this DecoratedRuntimeEnv.  # noqa: E501
        :type: str
        """

        self._pip_packages = pip_packages

    @property
    def conda_env_name(self):
        """Gets the conda_env_name of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The conda_env_name of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: str
        """
        return self._conda_env_name

    @conda_env_name.setter
    def conda_env_name(self, conda_env_name):
        """Sets the conda_env_name of this DecoratedRuntimeEnv.


        :param conda_env_name: The conda_env_name of this DecoratedRuntimeEnv.  # noqa: E501
        :type: str
        """

        self._conda_env_name = conda_env_name

    @property
    def conda_values(self):
        """Gets the conda_values of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The conda_values of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: object
        """
        return self._conda_values

    @conda_values.setter
    def conda_values(self, conda_values):
        """Sets the conda_values of this DecoratedRuntimeEnv.


        :param conda_values: The conda_values of this DecoratedRuntimeEnv.  # noqa: E501
        :type: object
        """

        self._conda_values = conda_values

    @property
    def env_variables(self):
        """Gets the env_variables of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The env_variables of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: dict(str, str)
        """
        return self._env_variables

    @env_variables.setter
    def env_variables(self, env_variables):
        """Sets the env_variables of this DecoratedRuntimeEnv.


        :param env_variables: The env_variables of this DecoratedRuntimeEnv.  # noqa: E501
        :type: dict(str, str)
        """

        self._env_variables = env_variables

    @property
    def created_at(self):
        """Gets the created_at of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The created_at of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: datetime
        """
        return self._created_at

    @created_at.setter
    def created_at(self, created_at):
        """Sets the created_at of this DecoratedRuntimeEnv.


        :param created_at: The created_at of this DecoratedRuntimeEnv.  # noqa: E501
        :type: datetime
        """
        if self.local_vars_configuration.client_side_validation and created_at is None:  # noqa: E501
            raise ValueError("Invalid value for `created_at`, must not be `None`")  # noqa: E501

        self._created_at = created_at

    @property
    def creator_id(self):
        """Gets the creator_id of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The creator_id of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: str
        """
        return self._creator_id

    @creator_id.setter
    def creator_id(self, creator_id):
        """Sets the creator_id of this DecoratedRuntimeEnv.


        :param creator_id: The creator_id of this DecoratedRuntimeEnv.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and creator_id is None:  # noqa: E501
            raise ValueError("Invalid value for `creator_id`, must not be `None`")  # noqa: E501

        self._creator_id = creator_id

    @property
    def creator(self):
        """Gets the creator of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The creator of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: MiniUser
        """
        return self._creator

    @creator.setter
    def creator(self, creator):
        """Sets the creator of this DecoratedRuntimeEnv.


        :param creator: The creator of this DecoratedRuntimeEnv.  # noqa: E501
        :type: MiniUser
        """
        if self.local_vars_configuration.client_side_validation and creator is None:  # noqa: E501
            raise ValueError("Invalid value for `creator`, must not be `None`")  # noqa: E501

        self._creator = creator

    @property
    def num_jobs(self):
        """Gets the num_jobs of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The num_jobs of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: int
        """
        return self._num_jobs

    @num_jobs.setter
    def num_jobs(self, num_jobs):
        """Sets the num_jobs of this DecoratedRuntimeEnv.


        :param num_jobs: The num_jobs of this DecoratedRuntimeEnv.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_jobs is None:  # noqa: E501
            raise ValueError("Invalid value for `num_jobs`, must not be `None`")  # noqa: E501

        self._num_jobs = num_jobs

    @property
    def num_named_actors(self):
        """Gets the num_named_actors of this DecoratedRuntimeEnv.  # noqa: E501


        :return: The num_named_actors of this DecoratedRuntimeEnv.  # noqa: E501
        :rtype: int
        """
        return self._num_named_actors

    @num_named_actors.setter
    def num_named_actors(self, num_named_actors):
        """Sets the num_named_actors of this DecoratedRuntimeEnv.


        :param num_named_actors: The num_named_actors of this DecoratedRuntimeEnv.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and num_named_actors is None:  # noqa: E501
            raise ValueError("Invalid value for `num_named_actors`, must not be `None`")  # noqa: E501

        self._num_named_actors = num_named_actors

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, DecoratedRuntimeEnv):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, DecoratedRuntimeEnv):
            return True

        return self.to_dict() != other.to_dict()



import json
from shlex import quote
import subprocess
import sys
import tempfile
from typing import Any, List, Tuple

import click

from anyscale.api import get_api_client
from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.client.openapi_client.models.execute_interactive_command_options import (
    ExecuteInteractiveCommandOptions,
)
from anyscale.cluster_config import get_cluster_config
import anyscale.conf
from anyscale.project import get_project_id, get_project_session, load_project_or_throw
from anyscale.util import execution_log_name


from anyscale.cluster_config import (  # noqa: N812 isort: skip
    add_anyscale_exec_node_provider as _DO_NOT_USE_add_anyscale_node_provider,
)


def rsync(*args: Any, **kwargs: Any) -> None:
    from ray.autoscaler.sdk import rsync as ray_rsync

    ray_rsync(*args, **kwargs)


class ExecController(object):
    def __init__(self, api_client: DefaultApi = None):
        if api_client is None:
            api_client = get_api_client()
        self.api_client = api_client

    def anyscale_exec(
        self,
        session_name: str,
        screen: bool,
        tmux: bool,
        port_forward: Tuple[int],
        sync: bool,
        stop: bool,
        terminate: bool,
        commands: List[str],
    ) -> None:
        import ray.scripts.scripts as autoscaler_scripts

        if screen and tmux:
            raise click.ClickException(
                "--screen and --tmux cannot both be specified. Please only specify one."
            )

        session_name, session_id = self._get_session_name_and_id(session_name)

        # Create a placeholder session command ID
        session_command = self.api_client.execute_interactive_command_api_v2_sessions_session_id_execute_interactive_command_post(
            session_id=session_id,
            execute_interactive_command_options=ExecuteInteractiveCommandOptions(
                shell_command=" ".join(commands)
            ),
        ).result
        session_command_id = session_command.command_id
        directory_name = session_command.directory_name
        dns_address = session_command.dns_address

        remote_command = self._generate_remote_command(
            session_command_id, commands, directory_name, stop, terminate
        )

        if tmux:
            wrapped_cmd = [
                "tmux",
                "new",
                "-d",
                "-s",
                session_command_id,
                "bash",
                "-c",
                quote(remote_command + "exec bash"),
            ]
            remote_command = " ".join(wrapped_cmd)
        if screen:
            wrapped_cmd = [
                "screen",
                "-L",
                "-dm",
                "bash",
                "-c",
                quote(remote_command + "exec bash"),
            ]
            remote_command = " ".join(wrapped_cmd)

        cluster_config = get_cluster_config(session_name, self.api_client)
        cluster_config = _DO_NOT_USE_add_anyscale_node_provider(
            cluster_config, dns_address
        )

        with tempfile.NamedTemporaryFile(mode="w") as config_file:
            json.dump(cluster_config, config_file)
            config_file.flush()
            config_file_path = config_file.name

            # Rsync file mounts if sync flag is set
            if sync:
                rsync(
                    cluster_config, source=None, target=None, down=False,
                )

            command = [
                sys.executable,
                autoscaler_scripts.__file__,
                "exec",
                config_file_path,
                remote_command,
                "--cluster-name",
                "{}".format(cluster_config["cluster_name"]),
            ]
            for port in list(port_forward):
                command.extend(["-p", str(port)])
            command_lst = [c for c in command if c]

            subprocess.check_call(command_lst, env=anyscale.ANYSCALE_ENV)  # noqa: B1

        if tmux or screen:
            launched_in_mode = "tmux" if tmux else "screen"
            run_cmd = "tmux attach" if tmux else "screen -xRR"
            click.echo(
                "Command launched in {mode}, log into the cluster using `anyscale ssh {name}` and run `{run_cmd}` to check on the status.".format(
                    mode=launched_in_mode, name=session_name, run_cmd=run_cmd
                )
            )

    def _get_session_name_and_id(self, session_name: str) -> Tuple[str, str]:
        project_definition = load_project_or_throw()
        project_id = get_project_id(project_definition.root)
        session = get_project_session(project_id, session_name, self.api_client)

        return session.name, session.id

    def _generate_remote_command(
        self,
        session_command_id: str,
        commands: List[str],
        directory_name: str,
        stop: bool,
        terminate: bool,
    ) -> str:
        # Save the PID of the command so we can kill it later.
        shell_command_prefix = (
            "echo $$ > {execution_log_name}.pid; "
            "export ANYSCALE_HOST={anyscale_host}; "
            "export ANYSCALE_SESSION_COMMAND_ID={session_command_id}; ".format(
                execution_log_name=execution_log_name(session_command_id),
                anyscale_host=anyscale.conf.ANYSCALE_HOST,
                session_command_id=session_command_id,
            )
        )

        # Note(simon): This section is largely similar to the server side exec command but simpler.
        # We cannot just use the server command because we need to buffer the output to
        # user's terminal as well and handle interactivity.
        redirect_to_dev_null = "&>/dev/null"
        shell_command = shell_command_prefix + " ".join(commands)
        remote_command = (
            "touch {execution_log_name}.out; "
            "touch {execution_log_name}.err; "
            f"cd {directory_name}; "
            "script -q -e -f -c {shell_command} {execution_log_name}.out; "
            "echo $? > {execution_log_name}.status; "
            "ANYSCALE_HOST={anyscale_host} anyscale session "
            "upload_command_logs --command-id {session_command_id} {redirect_to_dev_null}; "
            "ANYSCALE_HOST={anyscale_host} anyscale session "
            "finish_command --command-id {session_command_id} {stop_cmd} {terminate_cmd} {redirect_to_dev_null}; ".format(
                directory_name=directory_name,
                execution_log_name=(execution_log_name(session_command_id)),
                anyscale_host=anyscale.conf.ANYSCALE_HOST,
                session_command_id=session_command_id,
                stop_cmd="--stop" if stop else "",
                terminate_cmd="--terminate" if terminate else "",
                shell_command=quote(shell_command),
                redirect_to_dev_null=redirect_to_dev_null,
            )
        )

        return remote_command



from typing import Any, Dict, Optional, Tuple

from click import ClickException
from openapi_client.rest import ApiException

from anyscale.client.openapi_client.api.default_api import DefaultApi
from anyscale.util import format_api_exception


def get_cloud_json_from_id(cloud_id: str, api_client: DefaultApi) -> Dict["str", Any]:
    try:
        cloud = api_client.get_cloud_api_v2_clouds_cloud_id_get(
            cloud_id=cloud_id
        ).result
    except ApiException:
        return {
            "error": {
                "cloud_id": cloud_id,
                "message": f"The cloud with id, {cloud_id} has been deleted. Please create a new cloud with `anyscale cloud setup`.",
            }
        }
    return {
        "id": cloud.id,
        "name": cloud.name,
        "provider": cloud.provider,
        "region": cloud.region,
        "credentials": cloud.credentials,
        "config": cloud.config,
    }


def get_cloud_id_and_name(
    api_client: DefaultApi,
    cloud_id: Optional[str] = None,
    cloud_name: Optional[str] = None,
) -> Tuple[str, str]:
    if cloud_id and cloud_name:
        raise ClickException(
            "Both '--cloud-id' and '--cloud-name' specified. Please only use one."
        )
    elif cloud_name:
        with format_api_exception(ApiException):
            resp_get_cloud = api_client.find_cloud_by_name_api_v2_clouds_find_by_name_post(
                cloud_name_options={"name": cloud_name}
            )
        cloud = resp_get_cloud.result
    elif cloud_id:
        with format_api_exception(ApiException):
            resp_get_cloud = api_client.get_cloud_api_v2_clouds_cloud_id_get(
                cloud_id=cloud_id
            )

        cloud = resp_get_cloud.result
    else:
        try:
            clouds = api_client.list_clouds_api_v2_clouds_get().results
        except ApiException as e:
            if e.status == 404:
                # No clouds
                raise ClickException(
                    'There are no clouds assigned to your account. Please create a cloud using "anyscale cloud setup".'
                )
            with format_api_exception(ApiException):
                raise e

        if len(clouds) > 1:
            raise ClickException(
                "Multiple clouds: {}\n"
                "Please specify the one you want to refer to using --cloud-name.".format(
                    [cloud.name for cloud in clouds]
                )
            )
        cloud = clouds[0]
    return cloud.id, cloud.name



import contextlib
import logging
import os
import sys
from types import ModuleType
from typing import Any, Dict, List, Optional, Tuple

from ray.autoscaler._private.command_runner import _with_environment_variables
from ray.autoscaler.command_runner import CommandRunnerInterface

from anyscale.autoscaler.docker_local import client


logger = logging.getLogger(__name__)


def with_bash(cmd: str) -> List[str]:
    return ["/bin/bash", "-c", "--", cmd]


class DockerLocalCommandRunner(CommandRunnerInterface):  # type: ignore
    def __init__(
        self, log_prefix: str, node_id: str, process_runner: ModuleType
    ) -> None:
        self.log_prefix = log_prefix
        self.node_id = node_id
        self.process_runner = process_runner
        self._home_is_cached = False
        self._home_cached = ""

    def run(
        self,
        cmd: Optional[str] = None,
        timeout: int = 120,
        exit_on_fail: bool = False,
        port_forward: Optional[List[Tuple[int, int]]] = None,
        with_output: bool = False,
        environment_variables: Optional[Dict[str, object]] = None,
        run_env: str = "auto",
        ssh_options_override_ssh_key: str = "",
        shutdown_after_run: bool = False,
    ) -> str:
        if environment_variables:
            cmd = _with_environment_variables(str(cmd), environment_variables)
        final_cmd = with_bash(str(cmd))

        logger.info(
            self.log_prefix + 'Running "{}" in container {}.'.format(cmd, self.node_id)
        )
        container = client().containers.get(self.node_id)
        # TODO(dmitri): Probably should configure args passed to exec_run more
        # carefully. Refine this later, if needed.
        exit_code, output = container.exec_run(final_cmd)
        if exit_code != 0:
            logger.error(
                self.log_prefix + "Docker exec command failed. See "
                "command output below."
            )
            print(output.decode())
            if exit_on_fail:
                sys.exit(1)
        # TODO(dmitri): Container stdout and stderr are printed to stdout.
        # Refine this later, if needed.
        print(output.decode())

        if with_output:
            out = str(output.decode().strip())
            return out
        else:
            return ""

    def run_rsync_up(
        self, source: str, target: str, options: Optional[Dict[str, Any]] = None
    ) -> None:
        """Rsync files up to the cluster node.

        Args:
            source (str): The (local) source directory or file.
            target (str): The (remote) destination path.
        """
        target = self._remote_path(target)
        self._docker_cp(source, target)

    def run_rsync_down(
        self, source: str, target: str, options: Optional[Dict[str, Any]] = None
    ) -> None:
        """Rsync files down from the cluster node.

        Args:
            source (str): The (remote) source directory or file.
            target (str): The (local) destination path.
        """
        source = self._remote_path(source)
        self._docker_cp(source, target)

    def _docker_cp(self, source: str, target: str) -> None:
        logger.info(self.log_prefix + f"Copying {source} to {target}")
        cmd = with_bash(f"docker cp {source} {target}")
        try:
            self.process_runner.check_call(cmd)  # type: ignore
        except Exception:
            logger.exception(self.log_prefix + "docker cp failed.")

    def _remote_path(self, path: str) -> str:
        return ":".join([self.node_id, self._expand_user(path)])

    def _expand_user(self, path: str) -> str:
        if path[0] == "~":
            return self._home + path[1:]
        else:
            return path

    @property
    def _home(self) -> str:
        if not self._home_is_cached:
            # Silence stdout from this.
            with open(os.devnull, "w") as devnull:
                with contextlib.redirect_stdout(devnull):
                    self._home_cached = self.run("printenv HOME", with_output=True)
            self._home_is_cached = True
        return self._home_cached

    def remote_shell_command_str(self) -> str:
        """Return the command the user can use to open a shell."""
        return f"docker exec -it {self.node_id} /bin/bash"



import json
import os
from typing import Dict

import click

import anyscale


CREDENTIALS_FILE = "~/.anyscale/credentials.json"


def _validate_credential(token: str) -> None:
    if not token.startswith("sss_"):
        url = anyscale.util.get_endpoint("/credentials")
        raise click.ClickException(
            "Your user credentials are invalid. Please go to "
            f"{url} and follow the instructions to properly set your credentials."
        )


def load_credentials() -> str:
    # The environment variable ANYSCALE_CLI_TOKEN can be used to
    # overwrite the credentials in ~/.anyscale/credentials.json
    env_token = os.environ.get("ANYSCALE_CLI_TOKEN")
    if env_token is not None:
        _validate_credential(env_token)
        return env_token
    path = os.path.expanduser(CREDENTIALS_FILE)
    if not os.path.exists(path):
        url = anyscale.util.get_endpoint("/credentials")
        host = anyscale.util.get_endpoint("")
        raise click.ClickException(
            "Credentials not found. You need to create an account at {} "
            "and then go to {} and follow the instructions.".format(host, url)
        )
    with open(path) as f:
        try:
            credentials: Dict[str, str] = json.load(f)
        except json.JSONDecodeError:
            msg = (
                "Unable to load user credentials.\n\nTip: Try creating your "
                "user credentials again by going to {} and "
                "following the instructions. If this does not work, "
                "please contact Anyscale support!".format(
                    anyscale.util.get_endpoint("/credentials")
                )
            )
            raise click.ClickException(msg)
    received_token = credentials.get("cli_token")
    if received_token is None:
        raise click.ClickException(
            "The credential file is not valid. Please regenerate it by following "
            "the instructions at {}".format(anyscale.util.get_endpoint("/credentials"))
        )
    _validate_credential(received_token)
    return received_token



from typing import Any, Dict, Mapping, Tuple
from unittest.mock import Mock, patch

import pytest

from anyscale.autoscaler.aws.node_provider import AnyscaleAWSNodeProvider
from anyscale.autoscaler.node_provider import AnyscaleInstanceManagerNodeProvider


@pytest.mark.parametrize(
    "node_provider_class_and_config",
    [
        (AnyscaleAWSNodeProvider, {"region": "us-west-2"}),
        (
            AnyscaleInstanceManagerNodeProvider,
            {
                "inner_provider": {
                    "region": "us-west-2",
                    "type": "aws",
                    "cli_token": "fake_token",
                    "host": "fake_host",
                    "cloud_id": "fake_cloud_id",
                }
            },
        ),
    ],
)
@pytest.mark.parametrize(
    "cluster_config",
    [
        pytest.param({"provider": {"aws_credentials": {}}}, id="has aws_credentials"),
        pytest.param(
            {"provider": {"inner_provider": {"aws_credentials": {}}}},
            id="has inner_provider",
        ),
    ],
)
def test_ensure_no_credentials_on_head_node_config(
    cluster_config: Mapping[str, Any],
    node_provider_class_and_config: Tuple[Any, Dict[str, Any]],
) -> None:
    node_provider_class, provider_config = node_provider_class_and_config
    with patch.multiple(
        "ray.autoscaler._private.aws.node_provider",
        make_ec2_client=Mock(return_value=None),
    ):
        node_provider = node_provider_class(provider_config, "fake_cluster_name")

    cleaned_config = node_provider.prepare_for_head_node(cluster_config)

    provider_config = cleaned_config.get("provider", {})

    assert "aws_credentials" not in provider_config
    assert "aws_credentials" not in provider_config.get("inner_provider", {})


def test_create_node_return() -> None:
    mock_ec2_client = Mock()
    mock_instance_id = "fake_instance_id"
    mock_instance_ip = "fake_ip"
    mock_ec2_instance = Mock()
    mock_ec2_instance.id = mock_instance_id
    mock_ec2_instance.public_ip_address = mock_instance_ip
    mock_ec2_instance.state_reason = {}
    mock_ec2_instance.state = {"Name": "Fake Name"}
    mock_ec2_client.create_instances = Mock(return_value=[mock_ec2_instance])
    provider_config = {"region": "us-west-2", "cache_stopped_nodes": False}
    with patch.multiple(
        "ray.autoscaler._private.aws.node_provider",
        make_ec2_client=Mock(return_value=mock_ec2_client),
    ):
        node_provider = AnyscaleAWSNodeProvider(provider_config, "fake_cluster_name")
        created_node = node_provider.create_node({"SubnetIds": "123"}, tags={}, count=1)

        assert created_node[mock_instance_id]["PublicIpAddress"] == mock_instance_ip



# type: ignore
"""
This file is copied from open source ray:
    ray/_private/conda.py
"""

import hashlib
import json
import logging
import os
from pathlib import Path
import subprocess
from typing import List, Optional, Tuple, Union


"""Utilities for conda.  Adapted from https://github.com/mlflow/mlflow."""

logger = logging.getLogger(__name__)

# Name of environment variable indicating a path to a conda installation. Ray
# will default to running "conda" if unset.
RAY_CONDA_HOME = "RAY_CONDA_HOME"


def get_conda_activate_commands(conda_env_name: str) -> List[str]:
    """
    Get a list of commands to run to silently activate the given conda env.
    """
    #  Check whether it's from conda pack
    activate_file = Path(conda_env_name) / "bin" / "activate"
    if activate_file.exists():
        return [f"source {str(activate_file)}"]
    #  Checking for newer conda versions
    if os.name != "nt" and ("CONDA_EXE" in os.environ or RAY_CONDA_HOME in os.environ):
        conda_path = get_conda_bin_executable("conda")
        activate_conda_env = [
            "source {}/../etc/profile.d/conda.sh".format(os.path.dirname(conda_path))
        ]
        activate_conda_env += ["conda activate {} 1>&2".format(conda_env_name)]

    else:
        activate_path = get_conda_bin_executable("activate")
        # in case os name is not 'nt', we are not running on windows. Introduce
        # bash command otherwise.
        if os.name != "nt":
            return ["source %s %s 1>&2" % (activate_path, conda_env_name)]
        else:
            return ["conda activate %s" % (conda_env_name)]
    return activate_conda_env


def get_conda_bin_executable(executable_name: str) -> str:
    """
    Return path to the specified executable, assumed to be discoverable within
    the 'bin' subdirectory of a conda installation.

    The conda home directory (expected to contain a 'bin' subdirectory) is
    configurable via the ``RAY_CONDA_HOME`` environment variable. If
    ``RAY_CONDA_HOME`` is unspecified, this method simply returns the passed-in
    executable name.
    """
    conda_home = os.environ.get(RAY_CONDA_HOME)
    if conda_home:
        return os.path.join(conda_home, "bin/%s" % executable_name)
    # Use CONDA_EXE as per https://github.com/conda/conda/issues/7126
    if "CONDA_EXE" in os.environ:
        conda_bin_dir = os.path.dirname(os.environ["CONDA_EXE"])
        return os.path.join(conda_bin_dir, executable_name)
    return executable_name


def _get_conda_env_name(conda_env_path: str) -> str:
    conda_env_contents = open(conda_env_path).read() if conda_env_path else ""
    return "ray-%s" % hashlib.sha1(conda_env_contents.encode("utf-8")).hexdigest()


def get_or_create_conda_env(
    conda_env_path: str, env_name: str = None, base_dir: Optional[str] = None
) -> str:
    """
    Given a conda YAML, creates a conda environment containing the required
    dependencies if such a conda environment doesn't already exist. Returns the
    name of the conda environment, which is based on a hash of the YAML.

    Args:
        conda_env_path: Path to a conda environment YAML file.
        env_name(str, optional): The name of the environment to create. It'll be
            generated based on the hash of the config if not provided.
        base_dir (str, optional): Directory to install the environment into via
            the --prefix option to conda create.  If not specified, will
            install into the default conda directory (e.g. ~/anaconda3/envs)
    Returns:
        The name of the env, or the path to the env if base_dir is specified.
            In either case, the return value should be valid to pass in to
            `conda activate`.
    """
    conda_path = get_conda_bin_executable("conda")
    try:
        exec_cmd([conda_path, "--help"], throw_on_error=False)
    except EnvironmentError:
        raise ValueError(
            f"Could not find Conda executable at {conda_path}. "
            "Ensure Conda is installed as per the instructions at "
            "https://conda.io/projects/conda/en/latest/"
            "user-guide/install/index.html. "
            "You can also configure Ray to look for a specific "
            f"Conda executable by setting the {RAY_CONDA_HOME} "
            "environment variable to the path of the Conda executable."
        )
    _, stdout, _ = exec_cmd([conda_path, "env", "list", "--json"])
    envs = json.loads(stdout)["envs"]
    if env_name is None:
        env_name = _get_conda_env_name(conda_env_path)
    if base_dir:
        env_name = f"{base_dir}/{env_name}"
        if env_name not in envs:
            logger.info("=== Creating conda environment %s ===", env_name)
            exec_cmd(
                [
                    conda_path,
                    "env",
                    "create",
                    "--file",
                    conda_env_path,
                    "--prefix",
                    env_name,
                ],
                stream_output=True,
            )
        return env_name
    else:
        env_names = [os.path.basename(env) for env in envs]
        if env_name not in env_names:
            logger.info("=== Creating conda environment %s ===", env_name)
            exec_cmd(
                [conda_path, "env", "create", "-n", env_name, "--file", conda_env_path],
                stream_output=True,
            )
        return env_name


class ShellCommandException(Exception):
    pass


def exec_cmd(
    cmd: List[str], throw_on_error: bool = True, stream_output: bool = False
) -> Union[int, Tuple[int, str, str]]:
    """
    Runs a command as a child process.

    A convenience wrapper for running a command from a Python script.

    Note on the return value: If stream_output is true, then only the exit code
    is returned. If stream_output is false, then a tuple of the exit code,
    standard output and standard error is returned.

    Args:
        cmd: the command to run, as a list of strings
        throw_on_error: if true, raises an Exception if the exit code of the
            program is nonzero
        stream_output: if true, does not capture standard output and error; if
            false, captures these, streams and returns them
    """
    if stream_output:
        child = subprocess.Popen(cmd, universal_newlines=True, stdin=subprocess.PIPE)
        child.communicate()
        exit_code = child.wait()
        if throw_on_error and exit_code != 0:
            raise ShellCommandException("Non-zero exitcode: %s" % (exit_code))
        return exit_code
    else:
        child = subprocess.Popen(
            cmd,
            stdout=subprocess.PIPE,
            stdin=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )
        (stdout, stderr) = child.communicate()
        exit_code = child.wait()
        if throw_on_error and exit_code != 0:
            raise ShellCommandException(
                "Non-zero exit code: %s\n\nSTDOUT:\n%s\n\nSTDERR:%s"
                % (exit_code, stdout, stderr)
            )
        return exit_code, stdout, stderr



# coding: utf-8

"""
    Managed Ray API

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)  # noqa: E501

    The version of the OpenAPI document: 0.1.0
    Generated by: https://openapi-generator.tech
"""


import pprint
import re  # noqa: F401

import six

from openapi_client.configuration import Configuration


class MiniBuild(object):
    """NOTE: This class is auto generated by OpenAPI Generator.
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    """
    Attributes:
      openapi_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    openapi_types = {
        'id': 'str',
        'revision': 'int',
        'status': 'BuildStatus',
        'application_template_name': 'str'
    }

    attribute_map = {
        'id': 'id',
        'revision': 'revision',
        'status': 'status',
        'application_template_name': 'application_template_name'
    }

    def __init__(self, id=None, revision=None, status=None, application_template_name=None, local_vars_configuration=None):  # noqa: E501
        """MiniBuild - a model defined in OpenAPI"""  # noqa: E501
        if local_vars_configuration is None:
            local_vars_configuration = Configuration()
        self.local_vars_configuration = local_vars_configuration

        self._id = None
        self._revision = None
        self._status = None
        self._application_template_name = None
        self.discriminator = None

        self.id = id
        self.revision = revision
        self.status = status
        self.application_template_name = application_template_name

    @property
    def id(self):
        """Gets the id of this MiniBuild.  # noqa: E501

        Server assigned unique identifier.  # noqa: E501

        :return: The id of this MiniBuild.  # noqa: E501
        :rtype: str
        """
        return self._id

    @id.setter
    def id(self, id):
        """Sets the id of this MiniBuild.

        Server assigned unique identifier.  # noqa: E501

        :param id: The id of this MiniBuild.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and id is None:  # noqa: E501
            raise ValueError("Invalid value for `id`, must not be `None`")  # noqa: E501

        self._id = id

    @property
    def revision(self):
        """Gets the revision of this MiniBuild.  # noqa: E501

        Auto incrementing version number for this Build  # noqa: E501

        :return: The revision of this MiniBuild.  # noqa: E501
        :rtype: int
        """
        return self._revision

    @revision.setter
    def revision(self, revision):
        """Sets the revision of this MiniBuild.

        Auto incrementing version number for this Build  # noqa: E501

        :param revision: The revision of this MiniBuild.  # noqa: E501
        :type: int
        """
        if self.local_vars_configuration.client_side_validation and revision is None:  # noqa: E501
            raise ValueError("Invalid value for `revision`, must not be `None`")  # noqa: E501

        self._revision = revision

    @property
    def status(self):
        """Gets the status of this MiniBuild.  # noqa: E501

             Status of the Build.      `pending` - Build operation is queued and has not started yet.     `in_progress` - Build operation is in progress.     `succeeded` - Build operation completed successfully.     `failed` - Build operation completed unsuccessfully.     `pending_cancellation` - Build operation is marked for cancellation.     `cancelled` - Build operation was cancelled before it completed.       # noqa: E501

        :return: The status of this MiniBuild.  # noqa: E501
        :rtype: BuildStatus
        """
        return self._status

    @status.setter
    def status(self, status):
        """Sets the status of this MiniBuild.

             Status of the Build.      `pending` - Build operation is queued and has not started yet.     `in_progress` - Build operation is in progress.     `succeeded` - Build operation completed successfully.     `failed` - Build operation completed unsuccessfully.     `pending_cancellation` - Build operation is marked for cancellation.     `cancelled` - Build operation was cancelled before it completed.       # noqa: E501

        :param status: The status of this MiniBuild.  # noqa: E501
        :type: BuildStatus
        """
        if self.local_vars_configuration.client_side_validation and status is None:  # noqa: E501
            raise ValueError("Invalid value for `status`, must not be `None`")  # noqa: E501

        self._status = status

    @property
    def application_template_name(self):
        """Gets the application_template_name of this MiniBuild.  # noqa: E501

        The name of the App Config this build belongs to  # noqa: E501

        :return: The application_template_name of this MiniBuild.  # noqa: E501
        :rtype: str
        """
        return self._application_template_name

    @application_template_name.setter
    def application_template_name(self, application_template_name):
        """Sets the application_template_name of this MiniBuild.

        The name of the App Config this build belongs to  # noqa: E501

        :param application_template_name: The application_template_name of this MiniBuild.  # noqa: E501
        :type: str
        """
        if self.local_vars_configuration.client_side_validation and application_template_name is None:  # noqa: E501
            raise ValueError("Invalid value for `application_template_name`, must not be `None`")  # noqa: E501

        self._application_template_name = application_template_name

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.openapi_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, MiniBuild):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        if not isinstance(other, MiniBuild):
            return True

        return self.to_dict() != other.to_dict()
